#include "esp_camera.h"
#include <WiFi.h>
#include "esp_timer.h"
#include "img_converters.h"
#include "Arduino.h"
#include "fb_gfx.h"
#include "soc/soc.h"
#include "soc/rtc_cntl_reg.h"
#include "esp_http_server.h"
#include <WiFiClientSecure.h>
#include <UniversalTelegramBot.h>
#include <ArduinoJson.h>
#include "SD_MMC.h"
#include "FS.h"

// Telegram settings
#define BOTtoken "7926373507:AAG9h5hhIs-ESltQCLkxYTDivehR0BR3ciA"

// Array of chat IDs
const String CHAT_IDS[] = {
  "948503373",
  "6657372892",
  "THIRD_CHAT_ID",
  "FOURTH_CHAT_ID",
  "FIFTH_CHAT_ID"
};
const int NUM_CHAT_IDS = 5;

const char* ssid = "staff";
const char* password = "SRGB2020";

#define FLASH_LED_PIN 4
#define PART_BOUNDARY "123456789000000000000987654321"

// Variables
bool isRecording = false;
File videoFile;

// Telegram
WiFiClientSecure clientTCP;
UniversalTelegramBot bot(BOTtoken, clientTCP);

// Camera pins
#define CAMERA_MODEL_AI_THINKER
#define PWDN_GPIO_NUM     32
#define RESET_GPIO_NUM    -1
#define XCLK_GPIO_NUM      0
#define SIOD_GPIO_NUM     26
#define SIOC_GPIO_NUM     27
#define Y9_GPIO_NUM       35
#define Y8_GPIO_NUM       34
#define Y7_GPIO_NUM       39
#define Y6_GPIO_NUM       36
#define Y5_GPIO_NUM       21
#define Y4_GPIO_NUM       19
#define Y3_GPIO_NUM       18
#define Y2_GPIO_NUM        5
#define VSYNC_GPIO_NUM    25
#define HREF_GPIO_NUM     23
#define PCLK_GPIO_NUM     22

static const char* _STREAM_CONTENT_TYPE = "multipart/x-mixed-replace;boundary=" PART_BOUNDARY;
static const char* _STREAM_BOUNDARY = "\r\n--" PART_BOUNDARY "\r\n";
static const char* _STREAM_PART = "Content-Type: image/jpeg\r\nContent-Length: %u\r\n\r\n";

httpd_handle_t stream_httpd = NULL;

// Function to broadcast message to all chat IDs
void broadcastMessage(const String& message) {
  for (int i = 0; i < NUM_CHAT_IDS; i++) {
    bot.sendMessage(CHAT_IDS[i], message, "");
    delay(100); // Small delay between messages to prevent flooding
  }
}

bool initSDCard() {
  if(!SD_MMC.begin("/sdcard", true)) {
    Serial.println("SD Card Mount Failed");
    return false;
  }
  
  uint8_t cardType = SD_MMC.cardType();
  if(cardType == CARD_NONE) {
    Serial.println("No SD Card attached");
    return false;
  }
  
  return true;
}

void showMenu(String chat_id) {
  String welcome = "ESP32-CAM Bot Commands:\n\n";
  welcome += "/start : Show this menu\n";
  welcome += "/capture : Capture and save an image\n"; // Новая команда
  welcome += "/record : Start recording (with flash)\n";
  welcome += "/stoprecord : Stop recording\n";
  
  bot.sendMessage(chat_id, welcome, "");
}

// Функция для захвата и сохранения снимка
void captureImage(String chat_id) {
  if (!initSDCard()) {
    broadcastMessage("Failed to initialize SD Card");
    return;
  }

  camera_fb_t * fb = esp_camera_fb_get(); // Получаем кадр с камеры
  if (!fb) {
    broadcastMessage("Camera capture failed");
    return;
  }

  String filename = "/capture_" + String(millis()) + ".jpg"; // Имя файла
  File file = SD_MMC.open(filename, FILE_WRITE);

  if (!file) {
    broadcastMessage("Failed to create file");
    esp_camera_fb_return(fb); // Освобождаем буфер камеры
    return;
  }

  file.write(fb->buf, fb->len); // Записываем данные в файл
  file.close();
  esp_camera_fb_return(fb); // Освобождаем буфер камеры

  broadcastMessage("Capture saved: " + filename);
}

void startRecording(String chat_id) {
  if (!initSDCard()) {
    broadcastMessage("Failed to initialize SD Card");
    return;
  }

  String filename = "/capture_" + String(millis()) + ".mjpeg";
  videoFile = SD_MMC.open(filename, FILE_WRITE);
  
  if (!videoFile) {
    broadcastMessage("Failed to create file");
    return;
  }

  isRecording = true;
  digitalWrite(FLASH_LED_PIN, HIGH); // Turn on flash
  broadcastMessage("Recording started with flash. File: " + filename);
}

void stopRecording(String chat_id) {
  if (isRecording) {
    videoFile.close();
    isRecording = false;
    digitalWrite(FLASH_LED_PIN, LOW); // Turn off flash
    broadcastMessage("Recording stopped and saved");
  }
}

static esp_err_t stream_handler(httpd_req_t *req) {
  camera_fb_t * fb = NULL;
  esp_err_t res = ESP_OK;
  size_t _jpg_buf_len = 0;
  uint8_t * _jpg_buf = NULL;
  char * part_buf[64];

  res = httpd_resp_set_type(req, _STREAM_CONTENT_TYPE);
  if(res != ESP_OK) {
    return res;
  }

  while(true) {
    fb = esp_camera_fb_get();
    if (!fb) {
      Serial.println("Camera capture failed");
      res = ESP_FAIL;
    } else {
      if(fb->width > 400) {
        if(fb->format != PIXFORMAT_JPEG) {
          bool jpeg_converted = frame2jpg(fb, 80, &_jpg_buf, &_jpg_buf_len);
          esp_camera_fb_return(fb);
          fb = NULL;
          if(!jpeg_converted) {
            Serial.println("JPEG compression failed");
            res = ESP_FAIL;
          }
        } else {
          _jpg_buf_len = fb->len;
          _jpg_buf = fb->buf;
        }
      }
      
      if (isRecording && videoFile) {
        videoFile.write(_jpg_buf, _jpg_buf_len);
      }
    }
    
    if(res == ESP_OK) {
      size_t hlen = snprintf((char *)part_buf, 64, _STREAM_PART, _jpg_buf_len);
      res = httpd_resp_send_chunk(req, (const char *)part_buf, hlen);
    }
    if(res == ESP_OK) {
      res = httpd_resp_send_chunk(req, (const char *)_jpg_buf, _jpg_buf_len);
    }
    if(res == ESP_OK) {
      res = httpd_resp_send_chunk(req, _STREAM_BOUNDARY, strlen(_STREAM_BOUNDARY));
    }
    if(fb) {
      esp_camera_fb_return(fb);
      fb = NULL;
      _jpg_buf = NULL;
    } else if(_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;
    }
    if(res != ESP_OK) {
      break;
    }
  }
  return res;
}

void startCameraServer() {
  httpd_config_t config = HTTPD_DEFAULT_CONFIG();
  config.server_port = 80;

  httpd_uri_t index_uri = {
    .uri       = "/",
    .method    = HTTP_GET,
    .handler   = stream_handler,
    .user_ctx  = NULL
  };
  
  if (httpd_start(&stream_httpd, &config) == ESP_OK) {
    httpd_register_uri_handler(stream_httpd, &index_uri);
  }
}

void handleNewMessages(int numNewMessages) {
  for (int i = 0; i < numNewMessages; i++) {
    String chat_id = bot.messages[i].chat_id;
    String text = bot.messages[i].text;
    
    // Check if the sender is authorized
    bool isAuthorized = false;
    for (int j = 0; j < NUM_CHAT_IDS; j++) {
      if (chat_id == CHAT_IDS[j]) {
        isAuthorized = true;
        break;
      }
    }
    
    if (!isAuthorized) {
      bot.sendMessage(chat_id, "Unauthorized user", "");
      continue;
    }
    
    if (text == "/start") {
      showMenu(chat_id);
      String ipAddress = WiFi.localIP().toString();
      bot.sendMessage(chat_id, "Stream available at: http://" + ipAddress, "");
    }
    else if (text == "/capture") { // Обработка команды /capture
      captureImage(chat_id);
    }
    else if (text == "/record") {
      if (!isRecording) {
        startRecording(chat_id);
      } else {
        bot.sendMessage(chat_id, "Already recording", "");
      }
    }
    else if (text == "/stoprecord") {
      if (isRecording) {
        stopRecording(chat_id);
      } else {
        bot.sendMessage(chat_id, "Not recording", "");
      }
    }
  }
}

void setup() {
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);
  
  Serial.begin(115200);
  Serial.setDebugOutput(false);
  
  pinMode(FLASH_LED_PIN, OUTPUT);
  digitalWrite(FLASH_LED_PIN, LOW);
  
  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sccb_sda = SIOD_GPIO_NUM;
  config.pin_sccb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;
  config.pixel_format = PIXFORMAT_JPEG;
  
  if(psramFound()) {
    config.frame_size = FRAMESIZE_VGA;
    config.jpeg_quality = 18;
    config.fb_count = 2;
  } else {
    config.frame_size = FRAMESIZE_VGA;
    config.jpeg_quality = 18;
    config.fb_count = 1;
  }
  
  // Camera init
  esp_err_t err = esp_camera_init(&config);
  if (err != ESP_OK) {
    Serial.printf("Camera init failed with error 0x%x", err);
    return;
  }

  // Camera settings
  sensor_t * s = esp_camera_sensor_get();
  s->set_brightness(s, 0);
  s->set_contrast(s, 0);
  s->set_saturation(s, 0);
  s->set_special_effect(s, 0);
  s->set_whitebal(s, 1);
  s->set_awb_gain(s, 1);
  s->set_wb_mode(s, 0);
  s->set_exposure_ctrl(s, 1);
  s->set_aec2(s, 0);
  s->set_ae_level(s, 0);
  s->set_aec_value(s, 300);
  s->set_gain_ctrl(s, 1);
  s->set_agc_gain(s, 0);
  s->set_gainceiling(s, (gainceiling_t)0);
  s->set_bpc(s, 0);
  s->set_wpc(s, 1);
  s->set_raw_gma(s, 1);
  s->set_lenc(s, 1);
  s->set_hmirror(s, 0);
  s->set_vflip(s, 1);
  s->set_dcw(s, 1);
  s->set_colorbar(s, 0);
  
  // Connect to WiFi
  WiFi.begin(ssid, password);
  clientTCP.setCACert(TELEGRAM_CERTIFICATE_ROOT);
  
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("");
  Serial.println("WiFi connected");
  
  // Start web server
  startCameraServer();
  
  // Send initial message with menu
  String ipAddress = WiFi.localIP().toString();
  String welcome = "ESP32-CAM Started!\n\n";
  welcome += "Stream available at: http://" + ipAddress + "\n\n";
  welcome += "Commands:\n";
  welcome += "/start : Show menu\n";
  welcome += "/capture : Capture and save an image\n";
  welcome += "/record : Start recording (with flash)\n";
  welcome += "/stoprecord : Stop recording";
  
  broadcastMessage(welcome);
}

void loop() {
  int numNewMessages = bot.getUpdates(bot.last_message_received + 1);
  if (numNewMessages > 0) {
    handleNewMessages(numNewMessages);
  }
  delay(100);
}
